/**
 * 
  BaseKeyboard.cpp Implementations of hid boot Keyboard for "Hobo-nicola keyboard adapter".
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.

  
  Included in hoboNicola 1.2.1.		Dec. 1. 2018.
  
*/

#include "hbn_defines.h"
#if !defined(ESP32)
#include "HID.h"
#endif
#include "esp32_ble_hid_keyboard.h"
#include "BaseKeyboard.h"

static uint8_t hid_led_state = 0;
static uint8_t hid_idle_value = 0;
static void hid_notify(uint8_t request, uint8_t data) {
	switch(request) {
	case HID_SET_REPORT:
		hid_led_state = data;
		break;
	case HID_SET_IDLE:
		hid_idle_value = data;
		break;
	};
}

uint8_t BaseKeyboard::get_hid_led_state() const {
	return hid_led_state;
}

uint8_t BaseKeyboard::get_hid_idle_value() const {
	return hid_idle_value;
}

bool BaseKeyboard::isScrLock() const {
	return (hid_led_state & KEY_LED_SCRLOCK) != 0;
}

BaseKeyboard::BaseKeyboard() {
#if !defined(__ESP32_BLE_HID_KEYBOARD__)
	append_hid_desc();
	HID().set_request_callback(hid_notify);
#else
	ESP32_BLE_HID_Keyaboard::getInstance()->set_led_callback(led_func);
#endif
	report.modifiers = 0;
	for(int i = 0; i < REPORT_KEYS; i++)
		report.keys[i] = 0;
	modifier_keys = 0;
	last_report_time = 0;
}

void BaseKeyboard::sendReport(bool idle) {
#if !defined(__ESP32_BLE_HID_KEYBOARD__)
	HID().SendReport(2, &report, sizeof(key_report_t));
#else
	ESP32_BLE_HID_Keyaboard::getInstance()->sendReport(&report);
#endif
	if (!idle)
		last_report_time = millis();
}

void BaseKeyboard::releaseAll() {
	memset(&report, 0, sizeof(key_report_t));
	sendReport();
}

void BaseKeyboard::modifier_change(uint8_t before, uint8_t after) {
	uint8_t change = before ^ after;
	modifier_keys = after;
	if (change & after) {
		modifier_pressed(after);
		report_press(0, after); // 
	} else
		report_release(0, after);
}

void BaseKeyboard::report_press(uint8_t key, uint8_t mod) {
	modifier_keys = mod;
	if (key != 0) {
		bool already = false;
		int empty_slot = -1;
		for(uint8_t i = 0; i < REPORT_KEYS; i++) {
			if (report.keys[i] == key)
				already = true;
			if (report.keys[i] == 0 && empty_slot < 0)
				empty_slot = i;
		}
		if (empty_slot < 0)	// error condition.
			return;			// should report it ? 
		if (!already)
			report.keys[empty_slot] = key;
	}
	report.modifiers = modifier_keys;
	sendReport();  
}

void BaseKeyboard::report_release(uint8_t key, uint8_t mod, bool use_mod) {
	if (key != 0) {
		for(uint8_t i = 0; i < REPORT_KEYS; i++) {
			if (report.keys[i] == key) {
				report.keys[i] = 0;
				break;
			}
		}
	}
// UHSライブラリの問題。
// こうしておかないと、CTRLと文字を同時に離したときCTRLキーがオンのままになる。
	if (!use_mod)
		report.modifiers = modifier_keys;
	else		
		report.modifiers = mod;
	sendReport();  
}

bool BaseKeyboard::isAllReleased() const {
	for(uint8_t i = 0; i < REPORT_KEYS; i++) {
		if (report.keys[i] != 0)
			return false;
	}
	return true;
}

void BaseKeyboard::idle() {
	if (last_report_time != 0 && (millis() - last_report_time) > 250) {
		//sendReport(true);
		report_release(0, modifier_keys);
		last_report_time = 0;
	}
}


void BaseKeyboard::stroke(uint8_t key, uint8_t mod) {
	if (key) {
		report_press(key, mod);
		report_release(key, mod, true);
	}
}

//
// chart to hid usage id
// 文字コード - 0x20 でテーブルを引くことで、hid usage id を得る。
// SHIFTが必要なのものは、0x80をOR。
// usage id が8ビットのものは、事前に始末する。
#define SHIFT 0x80
const uint8_t char_to_hid_table[] PROGMEM = { // 日本語キーボード専用
  0x2c,      //  ' '
  0x1e|SHIFT,    // !
  0x1f|SHIFT,    // "
  0x20|SHIFT,    // #
  0x21|SHIFT,    // $
  0x22|SHIFT,    // %
  0x23|SHIFT,    // &
  0x24|SHIFT,    // '
  0x25|SHIFT,    // (
  0x26|SHIFT,    // )
  0x34|SHIFT,    // *
  0x33|SHIFT,    // +
  0x36,          // ,
  0x2d,          // -
  0x37,          // .
  0x38,          // /
  0x27,          // 0
  0x1e,          // 1
  0x1f,          // 2
  0x20,          // 3
  0x21,          // 4
  0x22,          // 5
  0x23,          // 6
  0x24,          // 7
  0x25,          // 8
  0x26,          // 9
  0x34,   // :
  0x33,     // ;
  0x36|SHIFT,   // <
  0x2d|SHIFT,   // =
  0x37|SHIFT,   // >
  0x38|SHIFT,   // ?
  0x2f,     // @
  0x04|SHIFT,   // A
  0x05|SHIFT,   // B
  0x06|SHIFT,   // C
  0x07|SHIFT,   // D
  0x08|SHIFT,   // E
  0x09|SHIFT,   // F
  0x0a|SHIFT,   // G
  0x0b|SHIFT,   // H
  0x0c|SHIFT,   // I
  0x0d|SHIFT,   // J
  0x0e|SHIFT,   // K
  0x0f|SHIFT,   // L
  0x10|SHIFT,   // M
  0x11|SHIFT,   // N
  0x12|SHIFT,   // O
  0x13|SHIFT,   // P
  0x14|SHIFT,   // Q
  0x15|SHIFT,   // R
  0x16|SHIFT,   // S
  0x17|SHIFT,   // T
  0x18|SHIFT,   // U
  0x19|SHIFT,   // V
  0x1a|SHIFT,   // W
  0x1b|SHIFT,   // X
  0x1c|SHIFT,   // Y
  0x1d|SHIFT,   // Z
  0x30,     // [
  0x89,          // \   !!!!
  0x32,          // ]
  0x2e,     // ^
  0x87|SHIFT,   // _  !!!!
  0x2f|SHIFT,   // `
  0x04,          // a
  0x05,          // b
  0x06,          // c
  0x07,          // d
  0x08,          // e
  0x09,          // f
  0x0a,          // g
  0x0b,          // h
  0x0c,          // i
  0x0d,          // j
  0x0e,          // k
  0x0f,          // l
  0x10,          // m
  0x11,          // n
  0x12,          // o
  0x13,          // p
  0x14,          // q
  0x15,          // r
  0x16,          // s
  0x17,          // t
  0x18,          // u
  0x19,          // v
  0x1a,          // w
  0x1b,          // x
  0x1c,          // y
  0x1d,          // z
  0x30|SHIFT,   // {
  0x89|SHIFT,   // |    !!!!
  0x32|SHIFT,   // }
  0x2e|SHIFT    // ~
};

// 文字に対応するhid usage id を返す。
// Shift修飾が必要な文字はbit8を1に。
uint16_t BaseKeyboard::char_to_hid(uint8_t c) {
	if (c == '\\') 		return 0x89;
	else if (c == '|') 	return 0x189;
	else if (c == '_') 	return 0x187;
	else if (c < 0x20 || c > 0x7f) 
		return 0;
	uint8_t cc = PGM_READ_BYTE(&char_to_hid_table[c - 0x20]);
	if (cc & 0x80) {
		cc &= 0x7f;
		return (cc | 0x100);
	}
	return cc;
}

void BaseKeyboard::strokeChar(uint8_t c, uint8_t& mod) {
	uint16_t k = char_to_hid(c);
	if ((k & 0x100) && !(mod & MODIFIER_HID_SHIFT_MASK))
		mod |= MODIFIRE_HID_SHIFT_L;           
	else if (!(k & 0x100) && (mod & MODIFIER_HID_SHIFT_MASK))
		mod &= ~MODIFIER_HID_SHIFT_MASK;
	stroke(k, mod);
}

//
// p holds an address of string in pgm.
//
void BaseKeyboard::send_PGM_string(const uint8_t* p) {
	if (!p) return;
	uint8_t save_mod = modifier_keys;
	uint8_t mod = modifier_keys;
	bool hid_direct = false;
	for(;;) {
		uint8_t c = PGM_READ_BYTE(p++);
		if (c == 0)
			break;
		if (c == HID_DIRECT_PREFIX)
			hid_direct = true;
		if (!hid_direct)
			strokeChar(c, mod);
		else
			stroke(c, 0);
	}
	modifier_keys = save_mod;
}

void BaseKeyboard::append_hid_desc() {
#if !defined(__ESP32_BLE_HID_KEYBOARD__)
	static HIDSubDescriptor node(_hidReportDescriptor, sizeof(_hidReportDescriptor));
	HID().AppendDescriptor(&node);
#endif
}

