/**
 * 
  nicola_keyboard.cpp   Main functionalities of "Hobo-nicola keyboard adapter".
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.
  
    Included in hoboNicola 1.2.1.		Dec. 1. 2018.
  
*/

#include "nicola_keyboard.h"

#define MODFIRE_KEY_MASK (MODIFIRE_HID_CTRL_MASK | MODIFIRE_HID_ALT_MASK | MODIFIRE_HID_GUI_MASK)

NicolaKeyboard::NicolaKeyboard() : BaseKeyboard() {
	state = Initial_State;
	nicola_state(Init);
	settings =  load_settings();
	setup_mode = false;
	disable_nicola = false;
	repeat_key = 0;
	repeat_time = 0;
	decicated_oyakeys = false;
}

void NicolaKeyboard::key_press(uint8_t k, uint8_t mod) {
	if (!k) {
		report_press(k, mod);
		return;	
	}
	if (!isNicola() || disable_nicola) {
		state = Initial_State;
		report_press(k, mod);
		if ((k == HID_HIRAGANA && is_kana_to_scr_on()) || (k == HID_ZENHAN && is_kanji_toggle_scrlock()))
			stroke(HID_SCRLOCK, modifier_keys);	// make scrlock on
		return;
	} else if (isNicola()) {
		if ((is_eisu_to_scr_off() && k == HID_CAPSLOCK) || (k == HID_ZENHAN && is_kanji_toggle_scrlock()))
			stroke(HID_SCRLOCK, modifier_keys);	// make scrlock off
	}
	uint16_t m = 0;
	uint8_t key = pre_convert(k);
	if (!key)
		m = convert_keycode(k);
	else
		k = key;   
	if (key || m == 0 || (mod & MODFIRE_KEY_MASK)) {
		nicola_state(All_off);  // flush
		report_press(k, mod);
		return ;
	}
	
	if (is_self_repeat()) {
		if (repeat_key != m)
			repeat_time = 0;
		repeat_key = m;
		if (repeat_time == 0)
			repeat_time = millis() + repeat_delay;
	}
	if (IsOyayubiKey(m))
		nicola_state(Oyayubi_pressed, m);
	else
		nicola_state(Moji_pressed, m);    
}

void NicolaKeyboard::key_release(uint8_t k, uint8_t mod) {

	repeat_key = 0;
	repeat_time = 0;

	if (!isNicola() || k == 0) {
		state = Initial_State;
		report_release(k, mod);
		return;
	}
	uint16_t m = 0;
	uint8_t key = pre_convert(k);
	if (!key)
		m = convert_keycode(k);
	else
		k = key;    
	if (key != 0 || m == 0 || (mod & MODFIRE_KEY_MASK) || state == Initial_State) {
		report_release(k, mod);
		return;
	}
	nicola_state(Key_released, m);
}

void NicolaKeyboard::modifier_pressed(uint8_t mod) {
	repeat_key = 0;
	repeat_time = 0;
	if (mod & MODFIRE_KEY_MASK)
		nicola_state(All_off);
}

static unsigned long blink_timer = 0;
static const unsigned long blink_interval = 300;
void NicolaKeyboard::idle() {
	BaseKeyboard::idle();
	unsigned long now = millis();
	if (!setup_mode) {
		HBN_NICOLA_LED(isNicola());
		blink_timer = 0;
	} else {
		if (blink_timer == 0)
			blink_timer = now;
		else if (now - blink_timer > blink_interval) {
			TOGGLE_HBN_NICOLA_LED();
			blink_timer = now;
		}
	}
	if (event_time != 0 && (now > event_time)) {
		nicola_state(Time_out);
		event_time = 0;
	}
	
	if (is_self_repeat() && repeat_key && (repeat_time != 0) && (now > repeat_time)) {
		nicola_state(Key_repeat, repeat_key);
		repeat_time = now + repeat_interval;
	}
}

void NicolaKeyboard::moji_set(uint16_t param) {
	moji = param;
	moji_time = millis();
	event_time = moji_time + e_charTime;
	repeat_moji = 0;
	repeat_oyayubi = 0;
	immediate_output(moji);
}

void NicolaKeyboard::oyayubi_set(uint16_t param) {
	oyayubi = param;
	oyayubi_time = millis();
	event_time = oyayubi_time + e_oyaTime;
	repeat_moji = 0;
	repeat_oyayubi = 0;
}

void NicolaKeyboard::nicola_state(nicola_event_t e, uint16_t param) {
	unsigned long now = millis();
	switch(state) {
	case Initial_State:
		event_time = 0;
		moji_time = 0;
		oyayubi_time = 0;
		oyayubi = 0;
		moji = 0;
		repeat_moji = 0;
		repeat_oyayubi = 0;
		immediate = false;
		switch(e) {
		case Moji_pressed:
			moji_set(param);
			state = Character_State;
			break;
		case Oyayubi_pressed:
			oyayubi_set(param);
			state = Oyayubi_State;
			break;
		default:
			break;
		}
		break;
	case Character_State:
		switch(e) {
		case Init:
			state = Initial_State;
			break;
		case All_off:
		case Key_released:
			output();
			state = Initial_State;
			break;
		case Moji_pressed:
			output();
			moji_set(param);
			break;
		case Oyayubi_pressed:
			oyayubi_set(param);
			state = Char_Oya_State;
			break;
		case Time_out:
		case Key_repeat:
			repeat_moji = moji;
			output();
			state = Repeat_State;
			break;
		}
		break;
	case Oyayubi_State:
		switch(e) {
		case Init:
			state = Initial_State;
			break;
		case Moji_pressed:
			repeat_oyayubi = oyayubi;
			repeat_moji = param;
			output();
			state = Repeat_State;
			break;
		case All_off:
			output();
			state = Initial_State;
			break;
		case Key_released:
			if (param == oyayubi) {
				output();
				state = Initial_State;
			}
			break;
		case Oyayubi_pressed:
			output();
			oyayubi_set(param);
			break;
		case Time_out:
		case Key_repeat:
			if (!decicated_oyakeys) {
				repeat_oyayubi = oyayubi;
				output();
				state = Repeat_State;
			}
			break;
		}
		break;
	case Char_Oya_State:
		switch(e) {
		case Init:
			state = Initial_State;
			break;
		case All_off:
			output();
			state = Initial_State;
			break;
		case Oyayubi_pressed:
			output();
			oyayubi_set(param);
			state = Oyayubi_State;
			break;
		case Moji_pressed:
			if ((oyayubi_time - moji_time) < (now - oyayubi_time)) {
				output();   // 先行の文字と親指
				moji_set(param);
				state = Character_State;
			} else {
				uint16_t oya = oyayubi;
				oyayubi = 0;
				if (!immediate)
					output();   // 先行の文字だけ出す
				immediate = false;
//				oyayubi = oya;
//				moji = param;
				repeat_oyayubi = oya;
				repeat_moji = param;
				output();   // 後追いの文字と親指
				state = Repeat_State;
			}
			break;
		case Key_released:  // 文字キーが離されたタイミングに応じて単独打鍵とする。
			if ((moji == param) && (now - oyayubi_time < e_nicolaTime) && (oyayubi_time - moji_time) > (now - oyayubi_time)) {
				uint16_t oya = oyayubi;
				oyayubi = 0;
				if (!immediate)
					output();  
				moji = 0;
				immediate = false;
				oyayubi = oya;
				state = Oyayubi_State;
				break;
			}
			output();
			state = Initial_State;
			break;
		case Time_out:
		case Key_repeat:
			repeat_oyayubi = oyayubi;
			repeat_moji = moji;
			output();
			state = Repeat_State;
			break;			
		}
		break;
	case Repeat_State:
		switch(e) {
		case Key_repeat:
			output();
			break;
		case Oyayubi_pressed:
			oyayubi_set(param);
			moji = 0;
			state = Oyayubi_State;
			break;		
		case Moji_pressed:
			moji_set(param);
			state = Character_State;
			break;
		case Time_out:
			break;	// ignore time-out event.	
		default:
			state = Initial_State;
			break;
		}	
		break;
	}
}

void NicolaKeyboard::output() {
	if (repeat_moji || repeat_oyayubi) {
		moji = repeat_moji;
		oyayubi = repeat_oyayubi;
	}
	if (moji == 0 && oyayubi) {
		stroke(HIGHBYTE(oyayubi), modifier_keys);
		oyayubi = 0;
		return;
	}
	if (immediate) {
		immediate = false;
		if (oyayubi)
			sendBS();
		else
			return;
	}
	if (isShiftPressed() && oyayubi == 0)
		oyayubi = NID_SHIFT_KEY;
	if (LOWBYTE(moji) == NID_SETUP_KEY) {
		if (LOWBYTE(oyayubi) == NID_LEFT_OYAYUBI) {
			oyayubi = 0;
			setup_mode = true;
		} else if (oyayubi == 0)
			setup_mode = false;
		moji = 0;
		return;
	}
	if (setup_mode) {
		setup_options(LOWBYTE(moji));
		moji = 0;
		oyayubi = 0;
		return;
	}

	const uint8_t* p = get_output_data(moji, oyayubi);
	oyayubi = 0;
	moji = 0;
	if (p)
		send_PGM_string(p);
}

void NicolaKeyboard::immediate_output(uint16_t moji) {
	if (!is_immediate_output() || setup_mode || LOWBYTE(moji) == NID_SETUP_KEY)
		return;
	const uint8_t* p = get_output_data(moji, isShiftPressed() ? NID_SHIFT_KEY : 0);
	if (p) {
		send_PGM_string(p);
		immediate = true;
	}
}

void NicolaKeyboard::sendBS() {
	stroke(HID_BACKSPACE, 0);
}

//
// キーボード初期化失敗時などに、 LEDを点滅して無限にループする。
//
void NicolaKeyboard::error_blink(int period) {
	HBN_NICOLA_LED(0);
	HBN_ERROR_LED(0);
	for(;;) {
		HBN_NICOLA_LED(1); delay(period); HBN_NICOLA_LED(0);delay(period);
		HBN_ERROR_LED(1); delay(period);HBN_ERROR_LED(0); delay(period);
  }  
}

